@@ -5,6 +5,13 @@ import hashlib
 import logging
 import shutil
 import socket
+from typing import Any
+from typing import Callable
+from typing import cast
+from typing import Dict
+from typing import List
+from typing import Mapping
+from typing import Optional
 
 from cryptography.hazmat.primitives import serialization
 import josepy as jose
@@ -13,7 +20,8 @@ import pytz
 
 from acme import fields as acme_fields
 from acme import messages
-from acme.client import ClientBase  # pylint: disable=unused-import
+from acme.client import ClientBase
+from certbot import configuration
 from certbot import errors
 from certbot import interfaces
 from certbot import util
@@ -46,11 +54,12 @@ class Account:
             cross-machine migration scenarios.
 
         """
-        creation_dt = acme_fields.RFC3339Field("creation_dt")
-        creation_host = jose.Field("creation_host")
-        register_to_eff = jose.Field("register_to_eff", omitempty=True)
+        creation_dt: datetime.datetime = acme_fields.rfc3339("creation_dt")
+        creation_host: str = jose.field("creation_host")
+        register_to_eff: str = jose.field("register_to_eff", omitempty=True)
 
-    def __init__(self, regr, key, meta=None):
+    def __init__(self, regr: messages.RegistrationResource, key: jose.JWK,
+                 meta: Optional['Meta'] = None) -> None:
         self.key = key
         self.regr = regr
         self.meta = self.Meta(
@@ -63,7 +72,10 @@ class Account:
         try:
             hasher = hashlib.md5()
         except ValueError:
-            hasher = hashlib.new('md5', usedforsecurity=False) # type: ignore
+            # This cast + dictionary expansion is made to make mypy happy without the need of a
+            # "type: ignore" directive that will also require to disable the check on useless
+            # "type: ignore" directives when mypy is run on Python 3.9+.
+            hasher = hashlib.new('md5', **cast(Mapping[str, Any], {"usedforsecurity": False}))
 
         hasher.update(self.key.key.public_key().public_bytes(
             encoding=serialization.Encoding.PEM,
@@ -78,16 +90,16 @@ class Account:
         # account key (and thus its fingerprint) to be updated...
 
     @property
-    def slug(self):
+    def slug(self) -> str:
         """Short account identification string, useful for UI."""
         return "{1}@{0} ({2})".format(pyrfc3339.generate(
             self.meta.creation_dt), self.meta.creation_host, self.id[:4])
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return "<{0}({1}, {2}, {3})>".format(
             self.__class__.__name__, self.regr, self.id, self.meta)
 
-    def __eq__(self, other):
+    def __eq__(self, other: Any) -> bool:
         return (isinstance(other, self.__class__) and
                 self.key == other.key and self.regr == other.regr and
                 self.meta == other.meta)
@@ -96,23 +108,24 @@ class Account:
 class AccountMemoryStorage(interfaces.AccountStorage):
     """In-memory account storage."""
 
-    def __init__(self, initial_accounts=None):
+    def __init__(self, initial_accounts: Dict[str, Account] = None) -> None:
         self.accounts = initial_accounts if initial_accounts is not None else {}
 
-    def find_all(self):
+    def find_all(self) -> List[Account]:
         return list(self.accounts.values())
 
-    def save(self, account, client):
+    def save(self, account: Account, client: ClientBase) -> None:
         if account.id in self.accounts:
             logger.debug("Overwriting account: %s", account.id)
         self.accounts[account.id] = account
 
-    def load(self, account_id):
+    def load(self, account_id: str) -> Account:
         try:
             return self.accounts[account_id]
         except KeyError:
             raise errors.AccountNotFound(account_id)
 
+
 class RegistrationResourceWithNewAuthzrURI(messages.RegistrationResource):
     """A backwards-compatible RegistrationResource with a new-authz URI.
 
@@ -122,38 +135,39 @@ class RegistrationResourceWithNewAuthzrURI(messages.RegistrationResource):
        continue to write out this field for some time so older
        clients don't crash in that scenario.
     """
-    new_authzr_uri = jose.Field('new_authzr_uri')
+    new_authzr_uri: str = jose.field('new_authzr_uri')
+
 
 class AccountFileStorage(interfaces.AccountStorage):
     """Accounts file storage.
 
-    :ivar .IConfig config: Client configuration
+    :ivar certbot.configuration.NamespaceConfig config: Client configuration
 
     """
-    def __init__(self, config):
+    def __init__(self, config: configuration.NamespaceConfig) -> None:
         self.config = config
         util.make_or_verify_dir(config.accounts_dir, 0o700, self.config.strict_permissions)
 
-    def _account_dir_path(self, account_id):
+    def _account_dir_path(self, account_id: str) -> str:
         return self._account_dir_path_for_server_path(account_id, self.config.server_path)
 
-    def _account_dir_path_for_server_path(self, account_id, server_path):
+    def _account_dir_path_for_server_path(self, account_id: str, server_path: str) -> str:
         accounts_dir = self.config.accounts_dir_for_server_path(server_path)
         return os.path.join(accounts_dir, account_id)
 
     @classmethod
-    def _regr_path(cls, account_dir_path):
+    def _regr_path(cls, account_dir_path: str) -> str:
         return os.path.join(account_dir_path, "regr.json")
 
     @classmethod
-    def _key_path(cls, account_dir_path):
+    def _key_path(cls, account_dir_path: str) -> str:
         return os.path.join(account_dir_path, "private_key.json")
 
     @classmethod
-    def _metadata_path(cls, account_dir_path):
+    def _metadata_path(cls, account_dir_path: str) -> str:
         return os.path.join(account_dir_path, "meta.json")
 
-    def _find_all_for_server_path(self, server_path):
+    def _find_all_for_server_path(self, server_path: str) -> List[Account]:
         accounts_dir = self.config.accounts_dir_for_server_path(server_path)
         try:
             candidates = os.listdir(accounts_dir)
@@ -180,15 +194,16 @@ class AccountFileStorage(interfaces.AccountStorage):
             accounts = prev_accounts
         return accounts
 
-    def find_all(self):
+    def find_all(self) -> List[Account]:
         return self._find_all_for_server_path(self.config.server_path)
 
-    def _symlink_to_account_dir(self, prev_server_path, server_path, account_id):
+    def _symlink_to_account_dir(self, prev_server_path: str, server_path: str,
+                                account_id: str) -> None:
         prev_account_dir = self._account_dir_path_for_server_path(account_id, prev_server_path)
         new_account_dir = self._account_dir_path_for_server_path(account_id, server_path)
         os.symlink(prev_account_dir, new_account_dir)
 
-    def _symlink_to_accounts_dir(self, prev_server_path, server_path):
+    def _symlink_to_accounts_dir(self, prev_server_path: str, server_path: str) -> None:
         accounts_dir = self.config.accounts_dir_for_server_path(server_path)
         if os.path.islink(accounts_dir):
             os.unlink(accounts_dir)
@@ -197,7 +212,7 @@ class AccountFileStorage(interfaces.AccountStorage):
         prev_account_dir = self.config.accounts_dir_for_server_path(prev_server_path)
         os.symlink(prev_account_dir, accounts_dir)
 
-    def _load_for_server_path(self, account_id, server_path):
+    def _load_for_server_path(self, account_id: str, server_path: str) -> Account:
         account_dir_path = self._account_dir_path_for_server_path(account_id, server_path)
         if not os.path.isdir(account_dir_path): # isdir is also true for symlinks
             if server_path in constants.LE_REUSE_SERVERS:
@@ -216,17 +231,21 @@ class AccountFileStorage(interfaces.AccountStorage):
 
         try:
             with open(self._regr_path(account_dir_path)) as regr_file:
-                regr = messages.RegistrationResource.json_loads(regr_file.read())
+                # TODO: Remove cast when https://github.com/certbot/certbot/pull/9073 is merged.
+                regr = cast(messages.RegistrationResource,
+                            messages.RegistrationResource.json_loads(regr_file.read()))
             with open(self._key_path(account_dir_path)) as key_file:
-                key = jose.JWK.json_loads(key_file.read())
+                # TODO: Remove cast when https://github.com/certbot/certbot/pull/9073 is merged.
+                key = cast(jose.JWK, jose.JWK.json_loads(key_file.read()))
             with open(self._metadata_path(account_dir_path)) as metadata_file:
-                meta = Account.Meta.json_loads(metadata_file.read())
+                # TODO: Remove cast when https://github.com/certbot/certbot/pull/9073 is merged.
+                meta = cast(Account.Meta, Account.Meta.json_loads(metadata_file.read()))
         except IOError as error:
             raise errors.AccountStorageError(error)
 
         return Account(regr, key, meta)
 
-    def load(self, account_id):
+    def load(self, account_id: str) -> Account:
         return self._load_for_server_path(account_id, self.config.server_path)
 
     def save(self, account: Account, client: ClientBase) -> None:
@@ -269,7 +288,7 @@ class AccountFileStorage(interfaces.AccountStorage):
         except IOError as error:
             raise errors.AccountStorageError(error)
 
-    def delete(self, account_id):
+    def delete(self, account_id: str) -> None:
         """Delete registration info from disk
 
         :param account_id: id of account which should be deleted
@@ -286,17 +305,18 @@ class AccountFileStorage(interfaces.AccountStorage):
         if not os.listdir(self.config.accounts_dir):
             self._delete_accounts_dir_for_server_path(self.config.server_path)
 
-    def _delete_account_dir_for_server_path(self, account_id, server_path):
+    def _delete_account_dir_for_server_path(self, account_id: str, server_path: str) -> None:
         link_func = functools.partial(self._account_dir_path_for_server_path, account_id)
         nonsymlinked_dir = self._delete_links_and_find_target_dir(server_path, link_func)
         shutil.rmtree(nonsymlinked_dir)
 
-    def _delete_accounts_dir_for_server_path(self, server_path):
+    def _delete_accounts_dir_for_server_path(self, server_path: str) -> None:
         link_func = self.config.accounts_dir_for_server_path
         nonsymlinked_dir = self._delete_links_and_find_target_dir(server_path, link_func)
         os.rmdir(nonsymlinked_dir)
 
-    def _delete_links_and_find_target_dir(self, server_path, link_func):
+    def _delete_links_and_find_target_dir(self, server_path: str,
+                                          link_func: Callable[[str], str]) -> str:
         """Delete symlinks and return the nonsymlinked directory path.
 
         :param str server_path: file path based on server
@@ -311,8 +331,8 @@ class AccountFileStorage(interfaces.AccountStorage):
 
         # does an appropriate directory link to me? if so, make sure that's gone
         reused_servers = {}
-        for k in constants.LE_REUSE_SERVERS:
-            reused_servers[constants.LE_REUSE_SERVERS[k]] = k
+        for k, v in constants.LE_REUSE_SERVERS.items():
+            reused_servers[v] = k
 
         # is there a next one up?
         possible_next_link = True
@@ -362,6 +382,6 @@ class AccountFileStorage(interfaces.AccountStorage):
                     uri=regr.uri)
             regr_file.write(regr.json_dumps())
 
-    def _update_meta(self, account, dir_path):
+    def _update_meta(self, account: Account, dir_path: str) -> None:
         with open(self._metadata_path(dir_path), "w") as metadata_file:
             metadata_file.write(account.meta.json_dumps())
